use std::{cell::RefCell, collections::HashMap, rc::Rc};

/// re-export fontdb
pub use fontdb;

pub struct FontStore {
    fontdb: fontdb::Database,
    loaded: RefCell<HashMap<(fontdb::ID, u32), Rc<fontdue::Font>>>,
}

impl FontStore {
    pub fn new_with_system_fonts() -> Self {
        let mut fontdb = fontdb::Database::new();
        fontdb.load_system_fonts();

        // query for a sans-serif font from a preselected list
        let sans_serif_result = fontdb
            .query(&fontdb::Query {
                families: &[
                    fontdb::Family::Name("DejaVu Sans"),
                    fontdb::Family::Name("Liberation Sans"),
                    fontdb::Family::Name("FreeSans"),
                    fontdb::Family::Name("OpenSans"),
                    fontdb::Family::Name("Arial"),
                    fontdb::Family::SansSerif,
                ],
                ..Default::default()
            })
            .expect("Couldn't find any usable fonts on the system");

        // use the resulting font family for a sans-serif font family
        let finfo = fontdb
            .face(sans_serif_result)
            .expect("couldn't load information for font just returned from a query");
        let family = finfo.families.first().expect("font has no family").clone();
        fontdb.set_sans_serif_family(family.0);

        Self {
            fontdb,
            loaded: Default::default(),
        }
    }

    pub fn query_for_font(&self, q: &fontdb::Query) -> Option<Rc<fontdue::Font>> {
        let qr = self.fontdb.query(q)?;
        self.fontdb.with_face_data(qr, |data, index| {
            let mut loaded = self.loaded.borrow_mut();
            let font = loaded.entry((qr, index)).or_insert_with(|| {
                Rc::new(
                    fontdue::Font::from_bytes(
                        data,
                        fontdue::FontSettings {
                            collection_index: index,
                            ..Default::default()
                        },
                    )
                    .expect("couldn't open font"),
                )
            });
            font.clone()
        })
    }
}

#[cfg(test)]
mod test {
    use super::FontStore;
    #[test]
    fn test_font_store() {
        let fs = FontStore::new_with_system_fonts();

        // query for sans serif font and check that it's nonempty
        assert!(fs
            .fontdb
            .query(&fontdb::Query {
                families: &[fontdb::Family::SansSerif,],
                weight: fontdb::Weight::NORMAL,
                style: fontdb::Style::Normal,
                ..Default::default()
            })
            .is_some());
    }
}
