use kahlo::math::{PixelBox, PixelSideOffsets};
use std::ops::Add;

use super::{HorizontalAlignment, LayoutNodeAccess, SizePolicy, SizePolicy2D, VerticalAlignment};

/// Child arrangement calculator trait.
///
/// Layout happens in two steps: the arrange step, and the layout step. The arrangement step
/// calculates constraints (e.g. minimum sizes) and computes general size policies for each layout
/// node. The layout step then uses the computed constraints and size policies to compute a set of
/// pixel-space rects.
pub trait ArrangementCalculator {
    fn arrange_step(&self, node: LayoutNodeAccess) -> SizePolicy2D;
    fn layout_step(&self, node: LayoutNodeAccess, inside: PixelBox);
}

mod container;
pub use container::ContainerArrangement;
mod line;
pub use line::LineArrangement;
mod table;
pub use table::TableArrangement;

fn do_align(node: &LayoutNodeAccess, mut inside: PixelBox) -> PixelBox {
    let net_policy = node
        .with_state(|ns| ns.net_policy)
        .expect("do_align invoked with node that has no net_policy");
    if net_policy.width.slack_weight == 0 {
        let slack = (inside.width() - net_policy.width.desired as i32).max(0);
        let hshrink = match node.halign() {
            HorizontalAlignment::Left => PixelSideOffsets::new(0, slack, 0, 0),
            HorizontalAlignment::Centre => {
                PixelSideOffsets::new(0, slack / 2, 0, slack - (slack / 2))
            }
            HorizontalAlignment::Right => PixelSideOffsets::new(0, 0, 0, slack),
        };
        inside = inside.inner_box(hshrink);
    }
    if net_policy.height.slack_weight == 0 {
        let slack = (inside.height() - net_policy.height.desired as i32).max(0);
        let vshrink = match node.valign() {
            VerticalAlignment::Top => PixelSideOffsets::new(0, 0, slack, 0),
            VerticalAlignment::Centre => {
                PixelSideOffsets::new(slack / 2, 0, slack - (slack / 2), 0)
            }
            VerticalAlignment::Bottom => PixelSideOffsets::new(slack, 0, 0, 0),
        };
        inside = inside.inner_box(vshrink);
    }
    inside
}

/// Calculate the fit of a number of size policies inside a given length.
///
/// Returns the endpoints of each item.
fn do_fit(
    total: i32,
    policies: impl Iterator<Item = SizePolicy> + Clone,
) -> impl Iterator<Item = i32> {
    // first pass over children: collect sum total of minimum/desired sizes and/slack weights
    let policy_sum = policies.clone().reduce(SizePolicy::add).unwrap_or_default();
    let (sum_min, sum_desired, sum_slack) = (
        policy_sum.minimum as i32,
        policy_sum.desired as i32,
        policy_sum.slack_weight as i32,
    );

    if total < sum_min {
        todo!("not enough space to distribute all children")
    }

    let (fit_space, slack_space) = if sum_desired <= total {
        (sum_desired - sum_min, total - sum_desired)
    } else {
        (total - sum_min, 0)
    };

    let total_fit_units = sum_desired - sum_min;

    let (fit_per, fit_per_rem) = if total_fit_units > 0 {
        (fit_space / total_fit_units, fit_space % total_fit_units)
    } else {
        (0, 0)
    };

    let (slack_per, slack_per_rem) = if sum_slack > 0 {
        (slack_space / sum_slack, slack_space % sum_slack)
    } else {
        (0, 0)
    };

    let mut allocated_fit = 0;
    let mut allocated_fit_rem = 0;

    let mut allocated_slack = 0;
    let mut allocated_slack_rem = 0;

    let mut min_allocation = 0;

    policies.map(move |w| {
        min_allocation += w.minimum as i32;

        let fit_units = (w.desired - w.minimum) as i32;
        if fit_units > 0 {
            allocated_fit += fit_per * fit_units;
            allocated_fit_rem += fit_per_rem * fit_units;
            allocated_fit += allocated_fit_rem / total_fit_units;
            allocated_fit_rem %= total_fit_units;
        }

        let slack_units = w.slack_weight as i32;
        if slack_units > 0 {
            allocated_slack += slack_per * slack_units;
            allocated_slack_rem += slack_per_rem * slack_units;
            allocated_slack += allocated_slack_rem / sum_slack;
            allocated_slack_rem %= sum_slack;
        }

        min_allocation + allocated_fit + allocated_slack
    })
}

#[cfg(test)]
mod tests {
    use crate::layout::SizePolicy;

    use super::do_fit;

    #[test]
    fn distribute_2_even() {
        let policies = vec![SizePolicy::expanding(1), SizePolicy::expanding(1)];

        assert_eq!(
            do_fit(2, policies.clone().into_iter()).collect::<Vec<_>>(),
            vec![1, 2]
        );
        assert_eq!(
            do_fit(3, policies.clone().into_iter()).collect::<Vec<_>>(),
            vec![1, 3]
        );
    }

    #[test]
    fn distribute_3_even() {
        let policies = vec![
            SizePolicy::expanding(1),
            SizePolicy::expanding(1),
            SizePolicy::expanding(1),
        ];

        assert_eq!(
            do_fit(3, policies.clone().into_iter()).collect::<Vec<_>>(),
            vec![1, 2, 3]
        );
        assert_eq!(
            do_fit(4, policies.clone().into_iter()).collect::<Vec<_>>(),
            vec![1, 2, 4]
        );
        assert_eq!(
            do_fit(5, policies.clone().into_iter()).collect::<Vec<_>>(),
            vec![1, 3, 5]
        );
        assert_eq!(
            do_fit(6, policies.clone().into_iter()).collect::<Vec<_>>(),
            vec![2, 4, 6]
        );
        assert_eq!(
            do_fit(7, policies.clone().into_iter()).collect::<Vec<_>>(),
            vec![2, 4, 7]
        );
    }
}
