use super::{do_fit, ArrangementCalculator};
use crate::{
    layout::{cache::NodeState, LayoutNodeAccess, SizePolicy, SizePolicy2D},
    math::{PixelBox, PixelPoint, PixelSize},
};
use std::ops::Add;

#[derive(Clone, Debug)]
pub enum LineArrangement {
    Column,
    Row,
}

impl LineArrangement {
    fn is_column(&self) -> bool {
        match self {
            Self::Column => true,
            Self::Row => false,
        }
    }
}

impl ArrangementCalculator for LineArrangement {
    fn arrange_step(&self, node: LayoutNodeAccess) -> SizePolicy2D {
        if node.child_count() == 0 {
            return SizePolicy2D::new(node.width_policy, node.height_policy)
                + node.margin_as_policy();
        }

        let child_policies = node
            .child_iter()
            .map(|child| child.with_state(|s| s.net_policy).unwrap())
            .collect::<Vec<_>>();

        let cw_policy = child_policies
            .iter()
            .map(|v| v.width)
            .reduce(if self.is_column() {
                SizePolicy::max
            } else {
                SizePolicy::add
            })
            .unwrap();
        let ch_policy: SizePolicy = child_policies
            .iter()
            .map(|v| v.height)
            .reduce(if self.is_column() {
                SizePolicy::add
            } else {
                SizePolicy::max
            })
            .unwrap();
        SizePolicy2D::new(
            node.width_policy.max_preserve_slack(cw_policy),
            node.height_policy.max_preserve_slack(ch_policy),
        ) + node.margin_as_policy()
    }

    fn layout_step(&self, node: LayoutNodeAccess, inside: PixelBox) {
        let inside = super::do_align(&node, inside);
        // do the final children layouts
        node.with_state(|ns| {
            if Some(inside) != ns.area {
                ns.area = Some(inside);
                ns.needs_render = true;
            }
        });

        if node.child_count() == 0 {
            return;
        }

        // add margins
        let inside = inside.inner_box(node.margin);

        // expansion direction extraction lambda
        let ee = if self.is_column() {
            |ns: &mut NodeState| ns.net_policy.height
        } else {
            |ns: &mut NodeState| ns.net_policy.width
        };

        let policies = node.child_iter().map(|c| c.with_state(ee).unwrap());
        let mut last_offset = 0;
        let fit = do_fit(
            if self.is_column() {
                inside.height()
            } else {
                inside.width()
            },
            policies,
        );

        // XXX: do 'alignment' of child stack here via an offset?

        for (offset, child) in fit.zip(node.child_iter()) {
            let cbox = if self.is_column() {
                PixelBox::from_origin_and_size(
                    PixelPoint {
                        x: inside.min.x,
                        y: inside.min.y + last_offset,
                        ..Default::default()
                    },
                    PixelSize {
                        width: inside.width(),
                        height: (offset - last_offset),
                        ..Default::default()
                    },
                )
            } else {
                PixelBox::from_origin_and_size(
                    PixelPoint {
                        x: inside.min.x + last_offset,
                        y: inside.min.y,
                        ..Default::default()
                    },
                    PixelSize {
                        width: (offset - last_offset),
                        height: inside.height(),
                        ..Default::default()
                    },
                )
            };

            child.child_arrangement.layout_step(child, cbox);
            last_offset = offset;
        }
    }
}
