use kahlo::{
    colour::Colour,
    math::PixelBox,
    prelude::{BitmapAccess, KahloOps},
};

use crate::{layout::LayoutNode, theme::Theme, window::RenderFormat};

mod text;
pub use text::TextCache;

#[derive(Default)]
pub enum ColourChoice {
    Background,
    Foreground,
    Active,
    Panel,
    Border,
    Custom(Colour),
    #[default]
    Default,
}

impl ColourChoice {
    fn get(self, theme: &Theme, default_to: ColourChoice) -> Colour {
        match self {
            Self::Background => theme.background,
            Self::Foreground => theme.foreground,
            Self::Active => theme.active,
            Self::Panel => theme.panel,
            Self::Border => theme.border,
            Self::Custom(col) => col,
            Self::Default => default_to.get(theme, Self::Custom(Colour::RED)),
        }
    }
}

pub struct RenderTarget<'r, 'data: 'r> {
    theme: &'r Theme,
    target: &'r mut kahlo::BitmapMut<'data, RenderFormat>,
}

impl<'a, 'r: 'a, 'data: 'r> RenderTarget<'r, 'data> {
    pub(crate) fn new(
        theme: &'r Theme,
        target: &'r mut kahlo::BitmapMut<'data, RenderFormat>,
    ) -> Self {
        Self { theme, target }
    }

    pub fn with_node(&'a mut self, lnode: &'a LayoutNode) -> NodeRenderTarget<'a, 'r, 'data> {
        if !lnode.render_check() {
            return NodeRenderTarget::SkipRender("node does not need to be rerendered");
        }
        let Some(area) = lnode.render_area() else {
            return NodeRenderTarget::SkipRender(
                "attempting to render node with no render area set",
            );
        };

        NodeRenderTarget::Render {
            rt: self,
            lnode,
            area,
        }
    }
}

pub enum NodeRenderTarget<'l, 'r: 'l, 'data: 'r> {
    Render {
        rt: &'l mut RenderTarget<'r, 'data>,
        lnode: &'l LayoutNode,
        area: PixelBox,
    },
    SkipRender(&'static str),
}

impl<'l, 'r: 'l, 'data: 'r> NodeRenderTarget<'l, 'r, 'data> {
    pub fn is_skipped(&self) -> bool {
        match self {
            Self::Render { .. } => false,
            Self::SkipRender(_) => true,
        }
    }
}

impl<'l, 'r: 'l, 'data: 'r> NodeRenderTarget<'l, 'r, 'data> {
    pub fn finish(self) {}

    pub fn fill(mut self, cc: ColourChoice) -> Self {
        if let Self::Render { rt, area, .. } = &mut self {
            rt.target
                .fill_region(*area, cc.get(rt.theme, ColourChoice::Background));
        }
        self
    }

    pub fn draw_foreground_outline(self) -> Self {
        self
    }

    pub fn outline(mut self, width: usize, cc: ColourChoice) -> Self {
        if let Self::Render { rt, area, .. } = &mut self {
            rt.target
                .rectangle(*area, width, cc.get(rt.theme, ColourChoice::Border));
        }
        self
    }

    pub fn border(mut self) -> Self {
        if let Self::Render { rt, area, .. } = &mut self {
            rt.target
                .rectangle(*area, rt.theme.border_width, rt.theme.border);
        }
        self
    }

    pub fn simple_text(mut self, text_cache: &TextCache, cc: ColourChoice) -> Self {
        if let Self::Render { rt, area, .. } = &mut self {
            text_cache.with_rendered(|amap| {
                rt.target.fill_region_masked(
                    amap,
                    PixelBox::from_size(amap.size()),
                    area.min,
                    cc.get(rt.theme, ColourChoice::Foreground),
                    kahlo::colour::BlendMode::SourceAlpha,
                );
            });
        }
        self
    }
}
