use crate::{
    input::InputState,
    layout::{HorizontalAlignment, LayoutNode, LayoutNodeAccess, TableCell, VerticalAlignment},
    render::RenderTarget,
    ui::UIHandle,
    Component,
};
use kahlo::math::PixelSideOffsets;

mod button;
mod frame;
mod group;
mod label;
mod spacer;
mod text_edit;
pub use button::Button;
pub use frame::Frame;
pub use group::{FormGroup, PlainGroup};
pub use label::Label;
pub use spacer::Spacer;
pub use text_edit::TextEdit;

pub type EventHook<C> = Box<dyn Fn() -> Option<<C as Component>::Msg>>;

pub trait Widget<C: Component> {
    fn poll(&mut self, uih: &mut UIHandle, input_state: Option<&InputState>) -> Vec<C::Msg>;

    fn layout_node(&self) -> LayoutNodeAccess;
    fn layout_node_mut(&mut self) -> &mut LayoutNode;
    fn render<'r, 'data: 'r>(&self, target: &mut RenderTarget<'r, 'data>);
}

pub trait WidgetExt<C: Component>: Widget<C> {
    fn boxed(self) -> Box<dyn Widget<C>>
    where
        Self: 'static + Sized,
    {
        Box::new(self)
    }

    fn framed(self) -> Frame<C>
    where
        Self: 'static + Sized,
    {
        Frame::wrap_widget(self)
    }

    fn set_margins(&mut self, margins: PixelSideOffsets) {
        *self.layout_node_mut().margin_mut() = margins;
    }
    fn with_margins(mut self, margins: PixelSideOffsets) -> Self
    where
        Self: Sized,
    {
        self.set_margins(margins);
        self
    }

    fn set_halign(&mut self, halign: HorizontalAlignment) {
        self.layout_node_mut().set_halign(halign);
    }
    fn with_halign(mut self, halign: HorizontalAlignment) -> Self
    where
        Self: Sized,
    {
        self.set_halign(halign);
        self
    }

    fn set_valign(&mut self, valign: VerticalAlignment) {
        self.layout_node_mut().set_valign(valign);
    }
    fn with_valign(mut self, valign: VerticalAlignment) -> Self
    where
        Self: Sized,
    {
        self.set_valign(valign);
        self
    }

    fn clear_table_cell(&mut self) {
        self.layout_node_mut().set_table_cell(None);
    }
    fn set_table_cell(&mut self, cell: TableCell) {
        self.layout_node_mut().set_table_cell(Some(cell));
    }
    fn with_table_cell(mut self, cell: TableCell) -> Self
    where
        Self: Sized,
    {
        self.set_table_cell(cell);
        self
    }
}

impl<C: Component, W: ?Sized + Widget<C>> WidgetExt<C> for W {}
