use crate::{
    input::MouseButton,
    layout::{HorizontalAlignment, LayoutNode, LayoutNodeAccess, LayoutTreeNode, LeafLayoutNode, SizePolicy},
    prelude::*,
    render::{ColourChoice, RenderTarget, TextCache},
    ui::UIHandle,
};
use winit::keyboard::{Key, NamedKey};

pub enum CursorPosition {
    Start,
    Index(usize),
    End,
}

pub struct TextEdit<C: Component> {
    lnode: LayoutNode,
    cursor_node: LeafLayoutNode,
    text: TextCache,
    cursor_pos: CursorPosition,
    onchange: Option<Box<dyn Fn(&str) -> Option<C::Msg>>>,
    focused: bool,
}

impl<C: Component> TextEdit<C> {
    pub fn new(uih: &UIHandle) -> Self {
        let minheight = (uih.theme().ui_font_size * 1.5).round() as usize;
        let mut lnode = uih.new_layout_node();
        lnode
            .set_width_policy(SizePolicy::expanding(1))
            .set_height_policy(SizePolicy::fixed(minheight));

        let mut cursor_node = uih.new_layout_node();
        cursor_node
            .set_width_policy(SizePolicy::fixed(2))
            .set_height_policy(SizePolicy::fixed(minheight))
            .set_halign(HorizontalAlignment::Left);

        Self {
            lnode,
            cursor_node: LeafLayoutNode::new(cursor_node),
            text: TextCache::new_with_font_and_size(
                uih.theme().ui_font.clone(),
                uih.theme().ui_font_size,
            ),
            cursor_pos: CursorPosition::End,
            onchange: None,
            focused: false
        }
    }

    pub fn set_text(&mut self, to: &str) {
        self.text.update(to);
    }

    /*pub fn set_cursor_position(&mut self, to: CursorPosition) {
        
    }*/
}

impl<C: Component> LayoutTreeNode<()> for TextEdit<C> {
    fn child(&self, ndx: usize) -> Option<LayoutNodeAccess<'_>> {
        if ndx == 0 {
            Some(self.cursor_node.access())
        } else {
            None
        }
    }
    fn child_count(&self) -> usize {
        1
    }
    fn current_node(&self) -> &LayoutNode {
        &self.lnode
    }
}

impl<C: Component> Widget<C> for TextEdit<C> {
    fn poll(
        &mut self,
        _uih: &mut UIHandle,
        input_state: Option<&crate::input::InputState>,
    ) -> Vec<<C as Component>::Msg> {
        let Some(istate) = input_state else {
            return vec![];
        };

        if istate.is_focused(self.lnode.id()) {
            if let Some(kp) = &istate.keypress {
                match kp {
                    Key::Named(NamedKey::Backspace) => {
                        self.text.change(String::pop);
                        self.lnode.render_needed();
                    }
                    Key::Named(NamedKey::Delete) => {
                        self.text.change(String::pop);
                        self.lnode.render_needed();
                    }
                    Key::Named(NamedKey::Enter) => {
                        istate.clear_focus();
                    }
                    _ => {
                        if let Some(inp) = &istate.text_input {
                            self.text.change(|s| s.push_str(inp.as_str()));
                            self.lnode.render_needed();
                        }
                    }
                }
            }
        }
        let render_area = self.lnode.render_area().unwrap();
        if istate
            .mouse_status_in(render_area, MouseButton::Left)
            .is_click()
        {
            istate.focus_on(self.lnode.id());
            self.focused = true;
            println!("click offset: {:?}", istate.mouse.pos - render_area.min);

            let cbar_offset = self.text.size_hint().width;
            self.cursor_node.margin_mut().left = cbar_offset;
        }

        let was_focused = self.focused;
        self.focused = istate.is_focused(self.lnode.id());

        if self.focused != was_focused {
            self.cursor_node.render_needed();
            self.lnode.render_needed();
        }

        vec![]
    }

    fn layout_node(&self) -> LayoutNodeAccess {
        LayoutNodeAccess::new(self)
    }
    fn layout_node_mut(&mut self) -> &mut LayoutNode {
        &mut self.lnode
    }

    fn render<'r, 'data: 'r>(&self, target: &mut RenderTarget<'r, 'data>) {
        target
            .with_node(&self.lnode)
            .fill(ColourChoice::Background)
            .simple_text(&self.text, ColourChoice::Foreground);

        if self.focused {
            target.with_node(&self.cursor_node)
                .fill(ColourChoice::Foreground);
        }
    }
}
